using System;
using System.Drawing;
using System.Xml;
using System.Globalization;
using Zcu.Mve.Numerics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Module for curve rendering.
	/// This module renders input uniform data array of Point2D using polyLine or Bezier interpolation.
	/// </summary>
	[ModuleInfo("Miroslav Vavruka", 
		 "Module for curve rendering. This module renders input uniform data array of Point2D using polyLine or Bezier interpolation.", 
		 Assemblied = "2005-10-11", 
		 IconName="CurveRenderer.ico")]
	[PortInfo("Input", "UniformDataArray of Point2D for rendering.")]
	public class CurveRenderer : Zcu.Mve.Core.Module
	{

		#region Fields and enums for graph properties settings

		#region Default renderer settings

		internal const int def_max_x = 600;
		internal const int def_max_y = 400;
		internal const int def_x_segment_count = 10;
		internal const int def_y_segment_count = 10;
		internal const bool def_showXGrid = false;
		internal const bool def_showYGrid = false;
		internal const bool def_showXAxesLegend = false;
		internal const bool def_showYAxesLegend = false;
		internal const int def_decPlaceCountX = 1;
		internal const int def_decPlaceCountY = 1;
		internal const bool def_showGraphMarks = true;
		internal const bool def_showGraphLine = true;
		internal const int def_graphLineWidth = 2;
		internal const int def_graphMarkLineWidth = 3;
		internal const int def_graph_mark_size = 6;
		internal const bool def_showRectangle = false;
		internal const bool def_showIndexes = false;
		internal const Interpolation def_interpolation = Interpolation.PolyLine;
		internal const bool def_allowHighlighting = true;
		internal const int def_hMarksSize = 8;
		internal const int def_hMarksLineWidth = 5;

		internal Color def_graphLineColor;
		internal Color def_marksLineColor;
		internal Color def_axesLineColor;
		internal Color def_axesTextColor;
		internal Color def_hMarksColor;
		internal Font def_axisFont;
		internal Font def_indexFont;
		internal Font def_hMarksFont;

		#endregion

		/// <summary>
		/// Type of interpolation for curve rendering.
		/// </summary>
		internal enum Interpolation
		{
			PolyLine,
			Bezier
		} // Interpolation


		/// <summary>
		/// x-resolution of panel.
		/// </summary>
		internal int max_x = def_max_x;
		/// <summary>
		/// y-resolution of panel.
		/// </summary>
		internal int max_y = def_max_y;

		/// <summary>
		/// Count of x-segments.
		/// </summary>
		internal int x_segment_count = def_x_segment_count;
		/// <summary>
		/// Count of y-segments.
		/// </summary>
		internal int y_segment_count = def_y_segment_count;

		/// <summary>
		/// Show x grid.
		/// </summary>
		internal bool showXGrid = def_showXGrid;
		/// <summary>
		/// Show y grid.
		/// </summary>
		internal bool showYGrid = def_showYGrid;

		/// <summary>
		/// Show x axes legend.
		/// </summary>
		internal bool showXAxesLegend = def_showXAxesLegend;
		/// <summary>
		/// Show y axes legend.
		/// </summary>
		internal bool showYAxesLegend = def_showYAxesLegend;

		/// <summary>
		/// Count of the decimal places on x-axis.
		/// </summary>
		internal int decPlaceCountX = def_decPlaceCountX;
		/// <summary>
		/// Count of the decimal places on y-axis.
		/// </summary>
		internal int decPlaceCountY = def_decPlaceCountY;

		/// <summary>
		/// Color of axes line.
		/// </summary>
		internal Color axesLineColor;
		/// <summary>
		/// Color of axes text.
		/// </summary>
		internal Color axesTextColor;
		/// <summary>
		/// Font of axis labels.
		/// </summary>
		internal Font axisFont;

		/// <summary>
		/// Show graph markers.
		/// </summary>
		internal bool showGraphMarks = def_showGraphMarks;
		/// <summary>
		/// Show graphLine
		/// </summary>
		internal bool showGraphLine = def_showGraphLine;

		/// <summary>
		/// Graph line width.
		/// </summary>
		internal int graphLineWidth = def_graphLineWidth;
		/// <summary>
		/// Graph marker size.
		/// </summary>
		internal int graphMarkLineWidth = def_graphMarkLineWidth;
		/// <summary>
		/// Graph marker size.
		/// </summary>
		internal int graph_mark_size = def_graph_mark_size;

		/// <summary>
		/// Show border line.
		/// </summary>
		internal bool showRectangle = def_showRectangle;
		/// <summary>
		/// Show point's indexes.
		/// </summary>
		internal bool showIndexes = def_showIndexes;

		/// <summary>
		/// Color of graph line.
		/// </summary>
		internal Color graphLineColor;
		/// <summary>
		/// Color of markers.
		/// </summary>
		internal Color marksLineColor;

		/// <summary>
		/// Type of interpolation.
		/// </summary>
		internal Interpolation interpolation = def_interpolation;

		/// <summary>
		/// Indexes font.
		/// </summary>
		internal Font indexFont;

		/// <summary>
		/// Allow selected points higlighting.
		/// </summary>
		internal bool allowHighlighting = def_allowHighlighting;
		/// <summary>
		/// Color of highlighted marks.
		/// </summary>
		internal Color hMarksColor;
		/// <summary>
		/// Size of highlighted marks.
		/// </summary>
		internal int hMarksSize = def_hMarksSize;
		/// <summary>
		/// Line Width of highlighted marks.
		/// </summary>
		internal int hMarksLineWidth = def_hMarksLineWidth;
		/// <summary>
		/// Font of highlighted marks.
		/// </summary>
		internal Font hMarksFont;

		#endregion


		#region Xml constants for WriteData() and ReadData() methods

		// node constants
		private const string XML_MAX_X               	 = "canvasWidth";
		private const string XML_MAX_Y               	 = "canvasHeight";
		private const string XML_X_SEGMENT_COUNT     	 = "xSegmentCount";
		private const string XML_Y_SEGMENT_COUNT     	 = "ySegmentCount";
		private const string XML_SHOW_X_GRID         	 = "showXGrid";
		private const string XML_SHOW_Y_GRID         	 = "showYGrid";
		private const string XML_SHOW_X_AXES_LEGEND  	 = "showXAxesLegend";
		private const string XML_SHOW_Y_AXES_LEGEND  	 = "showYAxesLegend";
		private const string XML_DEC_PLACE_COUNT_X   	 = "decPlaceCountX";
		private const string XML_DEC_PLACE_COUNT_Y   	 = "decPlaceCountY";
		private const string XML_AXES_LINE_COLOR     	 = "axesLineColor";
		private const string XML_AXES_TEXT_COLOR     	 = "axesTextColor";
		private const string XML_AXES_FONT               = "axesFont";
		private const string XML_SHOW_GRAPH_MARKS        = "showGraphMarks";
		private const string XML_SHOW_GRAPH_LINE         = "showGraphLine";
		private const string XML_GRAPH_LINE_WIDTH        = "graphLineWidth";
		private const string XML_GRAPH_MARK_LINE_WIDTH   = "graphMarkLineWidth";
		private const string XML_GRAPH_MARK_SIZE         = "graphMarkSize";
		private const string XML_SHOW_RECTANGLE          = "showRectangle";
		private const string XML_SHOW_INDEXES            = "showIndexes";
		private const string XML_GRAPH_LINE_COLOR        = "graphLineColor";
		private const string XML_MARKS_LINE_COLOR        = "marksLineColor";
		private const string XML_INTERPOLATION           = "interpolation";
		private const string XML_INDEX_FONT              = "indexFont";
		private const string XML_ALLOW_HIGHLIGHTING      = "allowHighlighting";
		private const string XML_H_MARKS_COLOR           = "hMarksColor";
		private const string XML_H_MARKS_SIZE            = "hMarksSize";
		private const string XML_H_MARKS_LINE_WIDTH      = "hMarksLineWidth";
		private const string XML_H_MARKS_FONT            = "hMarksFont";

		// color constants
		private const string XML_COLOR_ALPHA             = "a";
		private const string XML_COLOR_RED               = "r";
		private const string XML_COLOR_GREEN             = "g";
		private const string XML_COLOR_BLUE              = "b";

		// font constants
		private const string XML_FONT_NAME               = "name";
		private const string XML_FONT_SIZE               = "size";
		private const string XML_FONT_UNITS              = "units";
		private const string XML_FONT_STYLE              = "style";
		private const string XML_FONT_GDI_CHARSET        = "GdiCharSet";
		private const string XML_FONT_GDI_VERTICAL_FONT  = "GdiVerticalFont";

		#endregion


		/// <summary>
		/// UniformDataArray of Point2D for rendering.
		/// </summary>
		internal UniformDataArray data;

		/// <summary>
		/// Panel for graph rendering.
		/// </summary>
		internal CurveRendererPanel panel;


		/// <summary>
		/// Creates the input port.
		/// </summary>
		public CurveRenderer()
		{
			// properties initialization by default values
			def_graphLineColor = Color.Blue;
			def_marksLineColor = Color.FromArgb(255, 255, 128, 64);
			def_axesLineColor = Color.Black;
			def_axesTextColor = Color.Black;
			def_hMarksColor = Color.Red;
			def_axisFont = new Font(FontFamily.GenericSansSerif, 8, FontStyle.Bold);
			def_indexFont = new Font(FontFamily.GenericSansSerif, 8, FontStyle.Regular);
			def_hMarksFont = new Font(FontFamily.GenericSansSerif, 8, FontStyle.Bold);

			graphLineColor = def_graphLineColor;
			marksLineColor = def_marksLineColor;
			axesLineColor = def_axesLineColor;
			axesTextColor = def_axesTextColor;
			hMarksColor = def_hMarksColor;
			
			axisFont = def_axisFont;
			indexFont = def_indexFont;
			hMarksFont = def_hMarksFont;

			this.AddInPort("Input", typeof (UniformDataArray));
		} // CurveRenderer()


		/// <summary>
		/// Executes the CurveRenderer functionality.
		/// </summary>
		public override void Execute()
		{
			// get data from input port (it must be UniformDataArray of Point2D)
			data = (UniformDataArray) this.GetInput("Input");      

			if (data.ElemType == typeof(Zcu.Mve.Visualization.Point2D))
			{	// correct input data type -> OK, run renderer

				// there can be only one instance of one renderer panel at the same time!
				if (panel != null)
				{
					panel.Close();
					panel.Dispose();
				}

				panel = new CurveRendererPanel(this);
				System.Threading.Thread renThread = new System.Threading.Thread(new System.Threading.ThreadStart(this.RunRenderer));
                renThread.IsBackground = true;
				renThread.Start();
			}
			else
			{	// incorrect input data type (array doesn't consist of Point2D)
				throw new MveException(this, "Input array must consist of Point2D!");
			}
		} // Execute()


		/// <summary>
		/// Show the Renderer panel and render the curve.
		/// </summary>
		public void RunRenderer()
		{
			panel.Render();
		}


		/// <summary>
		/// Writes configuration data into XML element.
		/// </summary>
		/// <param name="config">XmlElement to be filled with configuration data.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			// integer nodes
			ConfigAppendNode(config, XML_MAX_X, this.max_x.ToString());
			ConfigAppendNode(config, XML_MAX_Y, this.max_y.ToString());
			ConfigAppendNode(config, XML_X_SEGMENT_COUNT, this.x_segment_count.ToString());
			ConfigAppendNode(config, XML_Y_SEGMENT_COUNT, this.y_segment_count.ToString());
			ConfigAppendNode(config, XML_DEC_PLACE_COUNT_X, this.decPlaceCountX.ToString());
			ConfigAppendNode(config, XML_DEC_PLACE_COUNT_Y, this.decPlaceCountY.ToString());
			ConfigAppendNode(config, XML_GRAPH_LINE_WIDTH, this.graphLineWidth.ToString());
			ConfigAppendNode(config, XML_GRAPH_MARK_LINE_WIDTH, this.graphMarkLineWidth.ToString());
			ConfigAppendNode(config, XML_GRAPH_MARK_SIZE, this.graph_mark_size.ToString());
			ConfigAppendNode(config, XML_H_MARKS_SIZE, this.hMarksSize.ToString());
			ConfigAppendNode(config, XML_H_MARKS_LINE_WIDTH, this.hMarksLineWidth.ToString());

			// color nodes
			ConfigAppendColorNode(config, XML_AXES_LINE_COLOR, this.axesLineColor);
			ConfigAppendColorNode(config, XML_AXES_TEXT_COLOR, this.axesTextColor);
			ConfigAppendColorNode(config, XML_GRAPH_LINE_COLOR, this.graphLineColor);
			ConfigAppendColorNode(config, XML_MARKS_LINE_COLOR, this.marksLineColor);
			ConfigAppendColorNode(config, XML_H_MARKS_COLOR, this.hMarksColor);

			// boolean nodes
			ConfigAppendNode(config, XML_SHOW_X_GRID, this.showXGrid.ToString());
			ConfigAppendNode(config, XML_SHOW_Y_GRID, this.showYGrid.ToString());
			ConfigAppendNode(config, XML_SHOW_X_AXES_LEGEND, this.showXAxesLegend.ToString());
			ConfigAppendNode(config, XML_SHOW_Y_AXES_LEGEND, this.showYAxesLegend.ToString());
			ConfigAppendNode(config, XML_SHOW_GRAPH_MARKS, this.showGraphMarks.ToString());
			ConfigAppendNode(config, XML_SHOW_GRAPH_LINE, this.showGraphLine.ToString());
			ConfigAppendNode(config, XML_SHOW_RECTANGLE, this.showRectangle.ToString());
			ConfigAppendNode(config, XML_SHOW_INDEXES, this.showIndexes.ToString());
			ConfigAppendNode(config, XML_ALLOW_HIGHLIGHTING, this.allowHighlighting.ToString());

			// font nodes
			ConfigAppendFontNode(config, XML_AXES_FONT, this.axisFont);
			ConfigAppendFontNode(config, XML_INDEX_FONT, this.indexFont);
			ConfigAppendFontNode(config, XML_H_MARKS_FONT, this.hMarksFont);

			// other nodes
			ConfigAppendNode(config, XML_INTERPOLATION, this.interpolation.ToString());

		} // WriteConfig()


		#region Auxiliary methods for WriteConfig()

		/// <summary>
		/// Appends config childNode.
		/// </summary>
		/// <param name="config">Config node - parent.</param>
		/// <param name="nodeName">Name of child node.</param>
		/// <param name="innerText">InnerText of child node.</param>
		private void ConfigAppendNode(XmlElement config, string nodeName, string innerText)
		{
			XmlElement elem = config.OwnerDocument.CreateElement(nodeName);
			elem.InnerText = innerText;
			config.AppendChild(elem);
		} // ConfigAppendNode()

		/// <summary>
		/// Appends config childNode. The node contains color data (4 attributes A, R, G, B).
		/// </summary>
		/// <param name="config">Config node - parent.</param>
		/// <param name="nodeName">Name of child node.</param>
		/// <param name="color">Color.</param>
		private void ConfigAppendColorNode(XmlElement config, string nodeName, Color color)
		{
			XmlElement elem = config.OwnerDocument.CreateElement(nodeName);
			elem.SetAttribute(XML_COLOR_ALPHA, color.A.ToString());
			elem.SetAttribute(XML_COLOR_RED, color.R.ToString());
			elem.SetAttribute(XML_COLOR_GREEN, color.G.ToString());
			elem.SetAttribute(XML_COLOR_BLUE, color.B.ToString());
			config.AppendChild(elem);
		} // ConfigAppendColorNode()

		/// <summary>
		/// Appends config childNode. The node contains Font data (as attributes).
		/// </summary>
		/// <param name="config">Config node - parent.</param>
		/// <param name="nodeName">Name of child node.</param>
		/// <param name="font">Font.</param>
		private void ConfigAppendFontNode(XmlElement config, string nodeName, Font font)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;

			XmlElement elem = config.OwnerDocument.CreateElement(nodeName);
			elem.SetAttribute(XML_FONT_NAME, font.Name.ToString());
			elem.SetAttribute(XML_FONT_SIZE, font.Size.ToString(nfi));
			elem.SetAttribute(XML_FONT_UNITS, font.Unit.ToString());
			elem.SetAttribute(XML_FONT_GDI_CHARSET, font.GdiCharSet.ToString());
			elem.SetAttribute(XML_FONT_GDI_VERTICAL_FONT, font.GdiVerticalFont.ToString());
			elem.SetAttribute(XML_FONT_STYLE, font.Style.ToString());

			config.AppendChild(elem);
		} // ConfigAppendFontNode()

		#endregion


		/// <summary>
		/// Reads configuration data from config XmlElement.
		/// </summary>
		/// <param name="config">XmlElement containing configuration data.</param>
		public override void ReadConfig(XmlElement config)
		{
			foreach (XmlNode node in config.ChildNodes)
			{
				switch (node.Name)
				{
						// integer nodes
					case XML_MAX_X:
						ParseIntNode(node, ref this.max_x);
						break;
					case XML_MAX_Y:
						ParseIntNode(node, ref this.max_y);
						break;
					case XML_X_SEGMENT_COUNT:
						ParseIntNode(node, ref this.x_segment_count);
						break;
					case XML_Y_SEGMENT_COUNT:
						ParseIntNode(node, ref this.y_segment_count);
						break;
					case XML_DEC_PLACE_COUNT_X:
						ParseIntNode(node, ref this.decPlaceCountX);
						break;
					case XML_DEC_PLACE_COUNT_Y:
						ParseIntNode(node, ref this.decPlaceCountY);
						break;
					case XML_GRAPH_LINE_WIDTH:
						ParseIntNode(node, ref this.graphLineWidth);
						break;
					case XML_GRAPH_MARK_LINE_WIDTH:
						ParseIntNode(node, ref this.graphMarkLineWidth);
						break;
					case XML_GRAPH_MARK_SIZE:
						ParseIntNode(node, ref this.graph_mark_size);
						break;
					case XML_H_MARKS_SIZE:
						ParseIntNode(node, ref this.hMarksSize);
						break;
					case XML_H_MARKS_LINE_WIDTH:
						ParseIntNode(node, ref this.hMarksLineWidth);
						break;

						// color nodes
					case XML_AXES_LINE_COLOR:
						ParseColorNode(node, ref this.axesLineColor);
						break;
					case XML_AXES_TEXT_COLOR:
						ParseColorNode(node, ref this.axesTextColor);
						break;
					case XML_GRAPH_LINE_COLOR:
						ParseColorNode(node, ref this.graphLineColor);
						break;
					case XML_MARKS_LINE_COLOR:
						ParseColorNode(node, ref this.marksLineColor);
						break;
					case XML_H_MARKS_COLOR:
						ParseColorNode(node, ref this.hMarksColor);
						break;

						// boolean nodes
					case XML_SHOW_X_GRID:
						ParseBoolNode(node, ref this.showXGrid);
						break;
					case XML_SHOW_Y_GRID:
						ParseBoolNode(node, ref this.showYGrid);
						break;
					case XML_SHOW_X_AXES_LEGEND:
						ParseBoolNode(node, ref this.showXAxesLegend);
						break;
					case XML_SHOW_Y_AXES_LEGEND:
						ParseBoolNode(node, ref this.showYAxesLegend);
						break;
					case XML_SHOW_GRAPH_MARKS:
						ParseBoolNode(node, ref this.showGraphMarks);
						break;
					case XML_SHOW_GRAPH_LINE:
						ParseBoolNode(node, ref this.showGraphLine);
						break;
					case XML_SHOW_RECTANGLE:
						ParseBoolNode(node, ref this.showRectangle);
						break;
					case XML_SHOW_INDEXES:
						ParseBoolNode(node, ref this.showIndexes);
						break;
					case XML_ALLOW_HIGHLIGHTING:
						ParseBoolNode(node, ref this.allowHighlighting);
						break;

						// font nodes
					case XML_AXES_FONT:
						ParseFontNode(node, ref this.axisFont);
						break;
					case XML_INDEX_FONT:
						ParseFontNode(node, ref this.indexFont);
						break;
					case XML_H_MARKS_FONT:
						ParseFontNode(node, ref this.hMarksFont);
						break;

						// other nodes
					case XML_INTERPOLATION:
						try
						{
							this.interpolation = (Interpolation) Enum.Parse(typeof(Interpolation), node.InnerText, true);
						}
						catch (Exception)
						{
							ParseErrorDebugLn("Interpolation parsing error in the node: '" + node.Name + "'");
						}
						break;

				} // switch()
			} // foreach()
		} // ReadConfig()


		#region Auxiliary methods for ReadConfig()

		/// <summary>
		/// Parses XML node containing integer value as innerText.
		/// </summary>
		/// <param name="node">XML Node with integer value.</param>
		/// <param name="intValue">Integer variable to be set</param>
		private void ParseIntNode(XmlNode node, ref int intValue)
		{
			try
			{
				intValue = Int32.Parse(node.InnerText);
			}
			catch (Exception)
			{
				ParseErrorDebugLn("Integer parsing error in the node: '" + node.Name + "'");
			}
		} // ParseIntNode()

		/// <summary>
		/// Parse color XML node.
		/// </summary>
		/// <param name="node">XMLNode.</param>
		/// <param name="color">Color variable to be set.</param>
		private void ParseColorNode(XmlNode node, ref Color color)
		{
			try
			{
				XmlElement element = (XmlElement) node;
				int alpha, red, green, blue;			
				alpha = Int32.Parse(element.GetAttribute(XML_COLOR_ALPHA));
				red   = Int32.Parse(element.GetAttribute(XML_COLOR_RED));
				green = Int32.Parse(element.GetAttribute(XML_COLOR_GREEN));
				blue  = Int32.Parse(element.GetAttribute(XML_COLOR_BLUE));
				color = Color.FromArgb(alpha, red, green, blue);
			}
			catch (Exception)
			{
				ParseErrorDebugLn("Color parsing error in the node: '" + node.Name + "'");
			}
		} // ParseColorNode()

		/// <summary>
		/// Parse XML node containing logical value as innerText.
		/// </summary>
		/// <param name="node">Xml node with logical value.</param>
		/// <param name="boolValue">Boolean variable to be set.</param>
		private void ParseBoolNode(XmlNode node, ref bool boolValue)
		{
			try
			{
				boolValue = Boolean.Parse(node.InnerText);
			}
			catch (Exception)
			{
				ParseErrorDebugLn("Boolean parsing error in the node: '" + node.Name + "'");
			}
		} // ParseBoolNode()

		/// <summary>
		/// Parse XML node containing font.
		/// </summary>
		/// <param name="node">Xml node with font information.</param>
		/// <param name="font">Font to be set.</param>
		private void ParseFontNode(XmlNode node, ref Font font)
		{
			try
			{			
				NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
				XmlElement element = (XmlElement) node;

				string family = element.GetAttribute(XML_FONT_NAME);
				float size = Single.Parse(element.GetAttribute(XML_FONT_SIZE), nfi);
				GraphicsUnit unit = (GraphicsUnit) Enum.Parse(typeof(GraphicsUnit), element.GetAttribute(XML_FONT_UNITS), true);
				FontStyle style = (FontStyle) Enum.Parse(typeof(FontStyle), element.GetAttribute(XML_FONT_STYLE), true);
				byte gdiCharSet = Byte.Parse(element.GetAttribute(XML_FONT_GDI_CHARSET));
				bool gdiVerticalFont = Boolean.Parse(element.GetAttribute(XML_FONT_GDI_VERTICAL_FONT));				

				font = new Font(family, size, style, unit, gdiCharSet, gdiVerticalFont);
			}
			catch (Exception)
			{
				ParseErrorDebugLn("Font parsing error in the node: '" + node.Name + "'");
			}
		} // ParseFontNode()

		/// <summary>
		/// Writes parsing error message into debug console.
		/// </summary>
		/// <param name="message">Error message.</param>
		private void ParseErrorDebugLn(string message)
		{
			string fullMsg = "An error occurred while ReadConfig(): " + message;
			DebugLn(fullMsg);
		} // ParseErrorDebugLn()

		#endregion	


		/// <summary>
		/// Invoke model setup.
		/// Returns null, setup is accessible from renderer panel.
		/// </summary>
		/// <returns>Null.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return null;
		} // InvokeSetup()


	} // class CurveRenderer
} // namespace
