using System;
using System.Collections;
using System.Xml;
using System.Diagnostics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Two dimensional structured grid.
	/// Contains two tables:
	/// <ul>
	///  <li>table of cell (pixel) attributes with height x width resolution 2D uniform arrays.</li>
	///  <li>table of point ("pixel vertices") attributes with height+1 x width+1 resolution 2D uniform arrays.</li>
	///  </ul>
	/// </summary>
	[DataObject("Two dimensional structured grid. Contains two tables: <ul>  <li>table of cell (pixel) attributes with height x width resolution 2D uniform arrays.</li>  <li>table of point (pixel vertices) attributes with height+1 x width+1 resolution 2D uniform arrays.</li>  </ul>")]
	public abstract class Grid2D : Zcu.Mve.Visualization.StructGrid
	{
		private Hashtable pointAttrs = new Hashtable();
		private Hashtable cellAttrs = new Hashtable();
		private int width;
		private int height;

		/// <summary>
		/// Returns horizontal pixel-resolution of Grid.
		/// </summary>
		public int Width
		{
			get
			{
				return this.width;
			}
		}

		/// <summary>
		/// Returns vertical pixel-resolution of Grid.
		/// </summary>
		public int Height
		{
			get
			{
				return this.height;
			}
		}
    
		/// <summary>
		/// For XmlLoader use only. Use parametric constructor instead.
		/// </summary>
		public Grid2D()
		{

		} // Grid2D()

		/// <summary>
		/// Constructor - creates empty grid of given resolution
		/// </summary>
		/// <param name="height">Number of pixels in vertical direction.</param>
		/// <param name="width">Number of pixels in horizontal direction.</param>
		public Grid2D(int height, int width)
		{
			this.width = width;
			this.height = height;
		} // Grid2D()

		/// <summary>
		/// Add cell (pixel) attributes.
		/// Array dimensions have to agree with Grid2D dimensions.
		/// </summary>
		/// <param name="attrid">Localy unique ID. Name of attribute type is recommended as default value.</param>
		/// <param name="attrs">2D array of attributes.</param>
		public void AddCellAttrs(string attrid, UniformDataArray2D attrs)
		{
			if (attrs.Height != this.height || attrs.Width != this.width)
			{
				throw new MveException(this, "Cell attributes dimensions and Grid2D dimensions have to agree.");
			}

			if (this.cellAttrs.ContainsKey(attrid))
			{
				throw new MveException(this, "Grid2D already contains cell attribute id " + attrid + ".");
			}

			this.cellAttrs.Add(attrid, attrs);
		}
    
		/// <summary>
		/// Add point attributes.
		/// Array dimensions have to agree with Grid2D dimensions.
		/// </summary>
		/// <param name="attrid">Localy unique ID. Name of attribute type is recommended as default value.</param>
		/// <param name="attrs">2D array of attributes.</param>
		public void AddPointAttrs(string attrid, UniformDataArray2D attrs)
		{
			if (attrs.Height != (this.height + 1) || (attrs.Width != this.width + 1))
			{
				throw new MveException(this, "Point attributes dimensions and Grid2D dimensions haveto agree.");
			}

			if (this.pointAttrs.ContainsKey(attrid))
			{
				throw new MveException(this, "Grid2D already contains point attribute id " + attrid + ".");
			}

			this.pointAttrs.Add(attrid, attrs);
		}

		/// <summary>
		/// Return 2D-array of pixel attributes.
		/// </summary>
		/// <param name="attrid">Localy unique ID. Name of attribute type is recommended as default value.</param>
		/// <returns>2D-array of pixel attributes.</returns>
		public UniformDataArray2D GetCellAttrs(string attrid)
		{
			if (!this.cellAttrs.ContainsKey(attrid))
				throw new MveException(this, "Grid2D do not contain pixel attributes with name " + attrid + ".");

			return (UniformDataArray2D) this.cellAttrs[attrid];
		}  

		/// <summary>
		/// Return 2D-array of point attributes.
		/// </summary>
		/// <param name="attrid">Localy unique ID. Name of attribute type is recommended as default value.</param>
		/// <returns>2D-array of point attributes.</returns>
		public UniformDataArray2D GetPointAttrs(string attrid)
		{
			if (!this.pointAttrs.ContainsKey(attrid))
				throw new MveException(this, "Grid2D do not contain point attributes with name " + attrid + ".");

			return (UniformDataArray2D) this.pointAttrs[attrid];
		}

		/// <summary>
		/// Returns collection of point attribute ids.
		/// </summary>
		/// <returns>Collection of point attribute ids.</returns>
		public ICollection GetPointAttrIds()
		{
			return this.pointAttrs.Keys;
		}

		/// <summary>
		/// Returns collection of cell attribute ids.
		/// </summary>
		/// <returns>Collection of cell attribute ids.</returns>
		public ICollection GetCellAttrIds()
		{
			return this.cellAttrs.Keys;
		}

		/// <summary>
		/// Returns point in i, j position. 
		/// Point coordinates are given implicitly. Thus no points are stored in Grid2D data object.
		/// Point object is created and coordinates are calculated in implementation of getPoint function.
		/// </summary>
		/// <param name="i">Height index</param>
		/// <param name="j">Width index</param>
		/// <returns>Particular point</returns>
		public abstract Point2D GetPoint(int i, int j);
    
		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public override IDataObject DeepCopy()
		{      
			Grid2D copy = (Grid2D) Activator.CreateInstance(this.GetType());

			copy.width = this.width;
			copy.height = this.height;
      
			foreach (string cellAttrId in this.GetCellAttrIds())
			{
				copy.AddCellAttrs(cellAttrId, this.GetCellAttrs(cellAttrId));
			}

			foreach (string pointAttrId in this.GetPointAttrIds())
			{
				copy.AddCellAttrs(pointAttrId, this.GetPointAttrs(pointAttrId));
			}

			return copy;
		}  

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public override void ReadData(XmlTextReader xmlTextReader)
		{
			ReadHeader(xmlTextReader);
			while(xmlTextReader.Read())
			{
				ReadXmlNode(xmlTextReader);
			} // while
		}

		/// <summary>
		/// Reads width and height attributes.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		protected void ReadHeader(XmlTextReader xmlTextReader)
		{
			this.height = (int) Int32.Parse(xmlTextReader.GetAttribute("height").Trim());
			this.width = (int) Int32.Parse(xmlTextReader.GetAttribute("width").Trim());
		}

		/// <summary>
		/// Reads pointattrs and cellattrs nodes.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		protected void ReadXmlNode(XmlTextReader xmlTextReader)
		{
			UniformDataArray2D array;
			string attrId;
     
			if (xmlTextReader.NodeType == XmlNodeType.Element)
			{
				switch(xmlTextReader.Name)
				{
					/*
					case "transform":
						this.transform = (AbstractTransform)Activator.CreateInstance(Globals.FindType(xmlTextReader.GetAttribute("type").Trim()));
						this.transform.ReadData(xmlTextReader);
						break;
					*/
					case "pointattrs":	// naten atribut bod
						attrId = xmlTextReader.GetAttribute("attrid").Trim();
						array = new UniformDataArray2D();
						array.ReadData(xmlTextReader);
						this.AddPointAttrs(attrId, array);
						break;
					case "cellattrs":	// naten atribut bunk
						attrId = xmlTextReader.GetAttribute("attrid").Trim();
						array = new UniformDataArray2D();
						array.ReadData(xmlTextReader);
						this.AddCellAttrs(attrId, array);
						break;
					default:
						Trace.WriteLine("Uknown node: " + xmlTextReader.Name + ".");
						break;
				} // switch
			}      
		}

		/// <summary>
		/// Write all point attributes into XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		protected void WritePointAttrs(XmlTextWriter xmlTextWriter)
		{
			// Write point attributes
			foreach (DictionaryEntry entry in this.pointAttrs)
			{
				UniformDataArray2D array = (UniformDataArray2D) entry.Value;
				xmlTextWriter.WriteStartElement("pointattrs");
				xmlTextWriter.WriteAttributeString("attrid", entry.Key.ToString());
				array.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}
		}

		/// <summary>
		/// Write all cell attributes into XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		protected void WriteCellAttrs(XmlTextWriter xmlTextWriter)
		{
			// Write cell attributes
			foreach (DictionaryEntry entry in this.cellAttrs)
			{
				UniformDataArray2D array = (UniformDataArray2D) entry.Value;
				xmlTextWriter.WriteStartElement("cellattrs");
				xmlTextWriter.WriteAttributeString("attrid", entry.Key.ToString());
				array.WriteData(xmlTextWriter);
				xmlTextWriter.WriteEndElement();
			}
		}

		/// <summary>
		/// Writes resolution of grid into XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		protected void WriteHeader(XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteAttributeString("height", this.Height.ToString());
			xmlTextWriter.WriteAttributeString("width", this.Width.ToString());
		}
  
		/// <summary>
		/// Writes width, height, point attrs and cell attrs.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public override void WriteData(XmlTextWriter xmlTextWriter)
		{
			this.WriteHeader(xmlTextWriter);
			this.WritePointAttrs(xmlTextWriter);
			this.WriteCellAttrs(xmlTextWriter);
		}
  
		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public override bool CheckConsistence()
		{
			return true;
		}

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string str = "Visualization.Grid2D:\r\n";
			str += "  height = " + this.height + "\r\n";
			str += "  width = " + this.width + "\r\n";
      
			str += "Point attributes:\r\n";
			foreach (DictionaryEntry entry in this.pointAttrs)
			{
				UniformDataArray2D array = (UniformDataArray2D)entry.Value;
				str += "  ID: " + entry.Key + " Type: " + array.ElemType.FullName + "\r\n";
			}
			str += "Cell attributes:\r\n";
			foreach (DictionaryEntry entry in this.cellAttrs)
			{
				UniformDataArray2D array = (UniformDataArray2D)entry.Value;
				str += "  ID: " + entry.Key + " Type: " + array.ElemType.FullName + "\r\n";
			}      
			return str;
		}

	} // class Grid2D 
} // namespace