using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.Drawing;
using System.ComponentModel;
using System.Xml;

namespace Zcu.Mve.Visualization
{	
    [ModuleInfo("Libor Va", "A module that transforms a Scalar value into a ColorRGBA. Values lower or equal to zero will become C1, values greater or equal to one will become C2, values in between will become linear interpolations of C1 and C2. This module can be used with the ScalarAttr2ColorRGBA module as a palette.", IconName = "PaletteLinear.ico")]
    [PortInfo("Value", "Scalar value to be transformed into ColorRGBA")]
    [PortInfo("Color", "Resulting ColorRGBA value")]
	/// <summary>
	/// A module that transforms a Scalar value into a ColorRGBA. Values lower or equal to zero will become C1, values greater or equal to one will become C2, values in between will become linear interpolations of C1 and C2. This module can be used with the ScalarAttr2ColorRGBA module as a palette.
	/// </summary>
    public class PaletteLinear:Module
    {
        private Color c1 = Color.Red;
        private Color c2 = Color.Green;

        /// <summary>
        /// Color that will be used for values lower than zero, and as a blend for values between zero and one
        /// </summary>
        [Description("Color that will be used for values lower than zero, and as a blend for values between zero and one")]
        public Color C1
        {
            get { return (c1); }
            set
            {
                c1 = value;
            }
        }

        /// <summary>
        /// Color that will be used for values greater than one, and as a blend for values between zero and one
        /// </summary>
        [Description("Color that will be used for values greater than one, and as a blend for values between zero and one")]
        public Color C2
        {
            get { return (c2); }
            set
            {
                c2 = value;
            }
        }

        /// <summary>
        /// Default constructor, adds input and output ports.
        /// </summary>
        public PaletteLinear()
        {
            AddInPort("Value", typeof(Scalar));
            AddOutPort("Color", typeof(ColorRGBA));
        }

        private float lerp(int v1, int v2, float w)
        {
            float result = (v1 / 255.0f) * w + (v2 / 255.0f) * (1 - w);
            return (result);
        }

        /// <summary>
        /// Performs the conversion of given value to ColorRGBA using linear interpolation
        /// </summary>
        public override void Execute()
        {
            Scalar input = (Scalar)GetInput("Value");
            float v = (float)input.Value;
            if (v < 0) 
                v = 0.0f;
            if (v > 1) 
                v = 1.0f;
            ColorRGBA output = new ColorRGBA(lerp(c1.R, c2.R, v), lerp(c1.G, c2.G, v), lerp(c1.B, c2.B, v));
            SetOutput("Color", output);
        }

        public override void WriteConfig(System.Xml.XmlElement config)
        {
            XmlElement ec1 = config.OwnerDocument.CreateElement("C1");
            ec1.InnerText = "" + c1.R + ", " + c1.G + ", " + c1.B;
            XmlElement ec2 = config.OwnerDocument.CreateElement("C2");
            ec2.InnerText = "" + c2.R + ", " + c2.G + ", " + c2.B;
            config.AppendChild(ec1);
            config.AppendChild(ec2);
        }

        public override void ReadConfig(XmlElement config)
        {
            XmlNodeList c1l = config.GetElementsByTagName("C1");
            if (c1l.Count == 0)
                c1 = Color.FromArgb(0, 255, 0);
            else
            {
                XmlNode c1n = c1l[0];
                string[] parts = c1n.InnerText.Split(new char[] { ',' });
                c1 = Color.FromArgb(int.Parse(parts[0]), int.Parse(parts[1]), int.Parse(parts[2]));
            }
            XmlNodeList c2l = config.GetElementsByTagName("C2");
            if (c2l.Count == 0)
                c2 = Color.FromArgb(255, 0, 0);
            else
            {
                XmlNode c2n = c2l[0];
                string[] parts = c2n.InnerText.Split(new char[] { ',' });
                c2 = Color.FromArgb(int.Parse(parts[0]), int.Parse(parts[1]), int.Parse(parts[2]));
            }
        }
    }
}
