using System;
using System.Drawing;
using System.Diagnostics;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Module for loading of 2D image from a file. Supported formats (JPEG, GIF, BMP). Image is returned as RegGrid2D.
	/// </summary>
	[ModuleInfo("Miroslav Vavruka and Milan Frank",
		 "Module for loading of 2D image from a file. Supported formats (JPEG, GIF, BMP). Image is returned as RegGrid2D.",
		 Assemblied = "2005-02-05",
		 IconName = "PictureLoader.ico")]
	[PortInfo("Output", "Loaded image.")]
	public class PictureLoader : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Name of the file containing data.
		/// </summary>
		private string fileName;

		/// <summary>
		/// Creates module and output port.
		/// </summary>
		public PictureLoader()
		{
			AddOutPort("Output", typeof (Visualization.RegGrid2D) );
			this.fileName = "";
		} // PictureLoader()

		/// <summary>
		/// Name of the file containing data.
		/// </summary>
		public string FileName
		{
			get
			{
				return this.fileName;
			}
			set
			{
				this.fileName = value;
			}
		} // FileName

		/// <summary>
		/// Read the configuration data from XML file.
		/// </summary>
		/// <param name="config">XmlElement containing the configuration data.</param>
		public override void ReadConfig(System.Xml.XmlElement config)
		{
			this.FileName = config.InnerText.Trim();
		} // ReadConfig()

		/// <summary>
		/// Write the configuration data in XML file.
		/// </summary>
		/// <param name="config">XmlElement containing the configuration data.</param>
		public override void WriteConfig(System.Xml.XmlElement config)
		{
			config.InnerText = this.FileName;
		} // WriteConfig()

		/// <summary>
		/// Executes PictureLoader functionality - load picture from the file.
		/// </summary>
		public override void Execute()
		{
			if (this.fileName == "")		// nezadano zadne jmeno souboru
				throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Pleas enter the name of file with data.");            
      
			try
			{
				Bitmap source;
				source = (Bitmap) Image.FromFile(this.fileName);
				UniformDataArray2D bmp = new UniformDataArray2D(typeof(ColorRGBA), source.Height, source.Width);

				for (int i = 0; i < bmp.Height; i++)
					for (int j = 0; j < bmp.Width; j++)
						bmp[i, j] = (ColorRGBA) source.GetPixel(j, i);
	
				RegGrid2D grid = new RegGrid2D(bmp.Height, bmp.Width);
				grid.AddCellAttrs(typeof(ColorRGBA).Name, bmp);
				SetOutput("Output", grid);
			} // try
			catch (OutOfMemoryException ex)
			{
				Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
				Debug.WriteLine(ex);
				throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Invalid file type. File must include a picture (*.bmp, *.jpg, *.gif etc.)");
			}
			catch (Exception ex)
			{
				Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
				Debug.WriteLine(ex);
				throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Could not find file \"" + ex.Message + "\"");
			} // catch
		} // Execute()

		/// <summary>
		/// Invoke module setup.
		/// </summary>
		/// <returns>Module setup dialog.</returns>
		public override ModuleSetup InvokeSetup()
		{
			return new PictureLoaderSetup(this);
		} // InvokeSetup()

	} // class PictureLoader
} // namespace
