using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// A module that produces a planar triangle mesh. The plane is represented by a point and two vectors of any dimension. The planes is split into specified number of facets. A, V1 and V2 should be vectors of equal dimension.
    /// </summary>
    [ModuleInfo("Libor Va", "A module that produces a planar triangle mesh. The plane is represented by a point and two vectors of any dimension. The planes is split into specified number of facets. A, V1 and V2 should be vectors of equal dimension.", IconName = "PlaneSource.ico")]
    [PortInfo("A","A point on the plane, on of the mesh corners.")]
    [PortInfo("V1", "One of the direction vectors of the plane. A+V1 will becone one of the corners of the plane.")]
    [PortInfo("V2", "One of the direction vectors of the plane. A+V2 will becone one of the corners of the plane.")]
	[PortInfo("Plane", "The resulting planar mesh.")]
    public class PlaneSource: Module
    {

        /// <summary>
        /// Default constructor, adds input and output ports.
        /// </summary>
        public PlaneSource()
        {
            AddInPort("V1", typeof(PointND));
            AddInPort("V2", typeof(PointND));
            AddInPort("A", typeof(PointND));
            AddOutPort("Plane", typeof(TriangleMesh));
        }

        private int xs = 10, ys = 10;

        private bool paramsChanged = true;

        /// <summary>
        /// Number of steps in the direction of the V1 vector
        /// </summary>
        [Description("Number of steps in the direction of the V1 vector, should be >=1.")]
        public int XS
        {
            get
            {
                return (xs);
            }
            set
            {
                if (value >= 1)
                {
                    xs = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Number of steps must be grater or equal to one.");
                }
            }
        }

        /// <summary>
        /// Number of steps in the direction of the V2 vector
        /// </summary>
        [Description("Number of steps in the direction of the V2 vector, should be >=1.")]
        public int YS
        {
            get
            {
                return (ys);
            }
            set
            {
                if (value >= 1)
                {
                    ys = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Number of steps must be grater or equal to one.");
                }
            }
        }

        private TriangleMesh result;

        // generates the resulting mesh, vertices are placed at positions A+x*V1+y*V2 where 0<=x<=1 and 0<=y<=1
        public override void Execute()
        {            
            if (paramsChanged || (!IsAllDataSame())||(result == null))
            {
                PointND v1 = (PointND)GetInput("V1");
                PointND v2 = (PointND)GetInput("V2");
                PointND a = (PointND)GetInput("A");

                result = new TriangleMesh();

                UniformDataArray points = new UniformDataArray(typeof(Point3D), xs * ys);

                double xst = 1.0 / xs;
                double yst = 1.0 / ys;

                for (int x = 0;x<xs;x++)
                    for (int y = 0; y < ys; y++)
                    {
                        points[x * ys + y] = (Point3D)(a + v1 * (x * xst) + v2 * (y*yst));
                    }

                UniformDataArray triangles = new UniformDataArray(typeof(Triangle), (xs - 1) * (ys - 1)*2);                

                for (int x = 0;x<(xs-1);x++)
                    for (int y = 0; y < (ys-1); y++)
                    {
                        int index = x * ys + y;
                        triangles[2 * (x*(ys-1)+y)] = new Triangle(index, index + ys, index + ys + 1);
                        triangles[2 * (x * (ys - 1) + y)+1] = new Triangle(index, index + ys + 1, index + 1);
                    }

                result.Points = points;
                result.AddCells("triangles", triangles);

                SetOutput("Plane", result);
                paramsChanged = false;
            }
            else
            {
                SetOutput("Plane", result, true);
            }
        }
    }
}
