using System;
using System.Globalization;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Represents a 2D point defined by two coordinates - X and Y.
	/// </summary>
	[DataObject("Represents a 2D point defined by two coordinates - X and Y.")]
	public struct Point2D : IPoint
	{
		/// <summary>
		/// x-coordinate of point.
		/// </summary>
		public double X;
		/// <summary>
		/// y-coordinate of point.
		/// </summary>
		public double Y;

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonDoubleImplicit;

		/// <summary>
		/// Create a Point2D object.
		/// </summary>
		/// <param name="x">x-coordinate of point.</param>
		/// <param name="y">y-coordinate of point.</param>
		public Point2D(double x, double y)
		{
			this.X = x;
			this.Y = y;
		} // Point2D()

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] pos = xmlTextReader.ReadString().Trim().Split(' ');
			this.X = Double.Parse(pos[0], nfi);
			this.Y = Double.Parse(pos[1], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteString(this.X.ToString(nfi) + " " + this.Y.ToString(nfi));
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new Point2D(this.X, this.Y);
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns the distance of point and coordinate origin.
		/// </summary>
		/// <returns>The distance of point from and coordinate origin.</returns>
		public double Abs()
		{
			return Math.Sqrt(this.X * this.X + this.Y * this.Y);
		} // Abs()

		/// <summary>
		/// Get/Set element of point.
		/// </summary>
		public double this [int index]
		{
			get
			{
				switch(index)
				{
					case 0:
						return this.X;
					case 1:
						return this.Y;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}

			set
			{
				switch(index)
				{
					case 0:
						this.X = value;
						break;
					case 1:
						this.Y = value;
						break;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
		}

		/// <summary>
		/// Get number of elements of point.
		/// </summary>
		public int Length
		{
			get
			{
				return 2;
			}
		}

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return "Point2D: [" + this.X + "; " + this.Y + "]";
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(Point2D) || b.GetType() != typeof(Point2D))
				return false;
		      
			return Math.Abs(((Point2D)a).X - ((Point2D)b).X) <= Point2D.Epsilon
				&& Math.Abs(((Point2D)a).Y - ((Point2D)b).Y) <= Point2D.Epsilon;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return Point2D.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			return this.X.GetHashCode() ^ this.Y.GetHashCode();
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversion

		// ===== Point2D -> Point =================
		/// <summary>
		/// Implicit conversion Point2D -> Point3D
		/// </summary>
		/// <param name="p">Point2D</param>
		/// <returns>Point3D</returns>
		public static implicit operator Point3D(Point2D p) 
		{
			return new Point3D(p.X, p.Y, 0);
		} // Point2D -> Point3D

		/// <summary>
		/// Implicit conversion Point2D -> PointND
		/// </summary>
		/// <param name="p">Point2D</param>
		/// <returns>PointND</returns>
		public static implicit operator PointND(Point2D p) 
		{
			return new PointND(new double[] {p.X, p.Y});
		} // Point2D -> PointND

		// ===== Point2D -> Vector ================
		/// <summary>
		/// Implicit conversion Point2D -> Vector2D
		/// </summary>
		/// <param name="p">Point2D</param>
		/// <returns>Vector2D</returns>
		public static implicit operator Zcu.Mve.Numerics.Vector2D(Point2D p) 
		{
			return new Zcu.Mve.Numerics.Vector2D(p.X, p.Y);
		} // Point2D -> Vector2D

		/// <summary>
		/// Implicit conversion Point2D -> Vector3D
		/// </summary>
		/// <param name="p">Point2D</param>
		/// <returns>Vector3D</returns>
		public static implicit operator Zcu.Mve.Numerics.Vector3D(Point2D p) 
		{
			return new Zcu.Mve.Numerics.Vector3D(p.X, p.Y, 0);
		} // Point2D -> Vector3D

		/// <summary>
		/// Implicit conversion Point2D -> VectorND
		/// </summary>
		/// <param name="p">Point2D</param>
		/// <returns>VectorND</returns>
		public static implicit operator Zcu.Mve.Numerics.VectorND(Point2D p) 
		{
			return new Zcu.Mve.Numerics.VectorND(new double[] {p.X, p.Y});
		} // Point2D -> VectorND

		// ===== Vector -> Point2D ================
		/// <summary>
		/// Implicit conversion Vector2D -> Point2D
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>Point2D</returns>
		public static implicit operator Point2D(Zcu.Mve.Numerics.Vector2D v) 
		{
			return new Point2D(v.X, v.Y);
		} // Vector2D -> Point2D

		/// <summary>
		/// Explicit conversion Vector3D -> Point2D
		/// </summary>
		/// <param name="v">Vector3D</param>
		/// <returns>Point2D</returns>
		public static explicit operator Point2D(Zcu.Mve.Numerics.Vector3D v) 
		{
			return new Point2D(v.X, v.Y);
		} // Vector3D -> Point2D

		/// <summary>
		/// Explicit conversion VectorND -> Point2D
		/// </summary>
		/// <param name="v">VectorND</param>
		/// <returns>Point2D</returns>
		public static explicit operator Point2D(Zcu.Mve.Numerics.VectorND v) 
		{
			if (v.Values.Length == 1)
				return new Point2D(v.Values[0], 0);
			else
				return new Point2D(v.Values[0], v.Values[1]);
		} // VectorND -> Point2D

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(Point2D a, Point2D b)
		{
			return Point2D.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(Point2D a, Point2D b) 
		{
			return !(a == b);
		}

        /// <summary>
        /// Operator of postmultiplication by double value. Performs per element multiplication.
        /// </summary>
        /// <param name="p">Left operand.</param>
        /// <param name="d">Right operand.</param>
        /// <returns>Result of multiplication.</returns>
        public static Point2D operator *(Point2D p, double d)
        {
            return(new Point2D(p.X*d, p.Y*d));
        }

        /// <summary>
        /// Operator of premultiplication by double value. Performs per element multiplication.
        /// </summary>
        /// <param name="d">Left operand.</param>
        /// <param name="p">Right operand.</param>
        /// <returns>Result of multiplication.</returns>
        public static Point2D operator *(double d, Point2D p)
        {
            return (new Point2D(p.X * d, p.Y * d));
        }

        /// <summary>
        /// Multiplication operator. Performs per element multiplication.
        /// </summary>
        /// <param name="p1">First argument.</param>
        /// <param name="p2">Second argument.</param>
        /// <returns>Result of multiplication.</returns>
        public static Point2D operator *(Point2D p1, Point2D p2)
        {
            return (new Point2D(p1.X * p2.X, p1.Y * p2.Y));
        }

        /// <summary>
        /// Division operator. Performs per element division.
        /// </summary>
        /// <param name="p1">First argument.</param>
        /// <param name="p2">Second argument.</param>
        /// <returns>Result of division.</returns>
        public static Point2D operator /(Point2D p1, Point2D p2)
        {
            return (new Point2D(p1.X / p2.X, p1.Y / p2.Y));
        }

        /// <summary>
        /// Operator of division by double value. Performs per element division.
        /// </summary>
        /// <param name="p">Point to be divided.</param>
        /// <param name="d">Divisor.</param>
        /// <returns>Result of division.</returns>
        public static Point2D operator /(Point2D p, double d)
        {
            return (new Point2D(p.X / d, p.Y / d));
        }

        /// <summary>
        /// Addition operator. Performs per element addition.
        /// </summary>
        /// <param name="p1">First argument.</param>
        /// <param name="p2">Second argument.</param>
        /// <returns>Resulting sum.</returns>
        public static Point2D operator +(Point2D p1, Point2D p2)
        {
            return (new Point2D(p1.X + p2.X, p1.Y + p2.Y));
        }

        /// <summary>
        /// Subtraction operator. Performs per element subtraction.
        /// </summary>
        /// <param name="p1">First argument.</param>
        /// <param name="p2">Second argument.</param>
        /// <returns>Difference.</returns>
        public static Point2D operator -(Point2D p1, Point2D p2)
        {
            return (new Point2D(p1.X - p2.X, p1.Y - p2.Y));
        }

		#endregion

	} // class Point2D
} // namespace
