using System;
using System.Globalization;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Represents a n-dimensional point defined by N coordinates.
	/// </summary>
	public struct PointND : IPoint
	{
		/// <summary>
		/// n-dimensional array of point coordinates.
		/// </summary>
		private double[] values;

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonFloatImplicit;

		/// <summary>
		/// Creates new instance of PointND of given dimension
		/// </summary>
		/// <param name="dimension">Dimension of PointND.</param>
		public PointND(int dimension)
		{
			if (dimension <= 0)
				throw new MveException(typeof(PointND),"PointND dimension must be at least one.");
			values = new double [dimension];
		}

		/// <summary>
		/// Create a PoinND object.
		/// </summary>
		/// <param name="values">n-dimensional array of vector values.</param>
		public PointND(params double[] values)
		{
			this.values = values;
		} // PointND()

		/// <summary>
		/// Creates new instance of PointND of given dimension and copies content of given point.
		/// </summary>
		/// <param name="point">Point to be copied</param>
		/// <param name="newDim">Dimension of new point</param>
		/// <returns></returns>
		public static PointND ChangeDimension(IPoint point, int newDim)
		{
			PointND ret = new PointND(newDim);

			if (point == null)
				return ret;

			for (int i = 0; i < ret.Length && i < point.Length; i++)
				ret[i] = point[i];

			return ret;
		}

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] pos = xmlTextReader.ReadString().Trim().Split(' ');
			this.values = new double[pos.Length];
			for (int i = 0; i < pos.Length; i++)
				this.Values[i] = Double.Parse(pos[i], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			string pom = "";
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			for (int i = 0; i < this.Values.Length; i++)
				pom += this.Values[i].ToString(nfi) + " ";
			pom = pom.Substring(0, pom.Length - 1);	// odstraneni prebytecne mezery na konci
			xmlTextWriter.WriteString(pom);
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			double[] newPosition = new double[this.Values.Length];
			this.Values.CopyTo(newPosition, 0);
			return new PointND(newPosition);
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns the distance of point and coordinate origin.
		/// </summary>
		/// <returns>The distance of point from and coordinate origin.</returns>
		public double Abs()
		{
			double tmp = 0;
			foreach (double d in this.values)
				tmp += Math.Pow(d, 2);
			return Math.Sqrt(tmp);
		} // Abs()

		/// <summary>
		/// Get the n-dimensional array of point coordinates.
		/// For acces to concrete point coordinates use indexer.
		/// </summary>
		public double[] Values
		{
			get
			{
				return this.values;
			}
		} // Values

		/// <summary>
		/// Get dimension of point.
		/// </summary>
		public int Length
		{
			get
			{
				return this.values.Length;
			}
		} // Length

		/// <summary>
		/// Dot product of two PointNDs
		/// </summary>
		/// <param name="p">secornd argument</param>
		/// <returns>Scalar value of the dot product</returns>
		public Scalar dotProduct(PointND p)
		{
			if (p.Length!=this.Length)
			{
				throw new MveException(this, "Cannot compute dot product of points of inequal dimension.");
			}
			else
			{
				double result = 0;
				for (int i = 0;i<this.Length;i++) result+= this[i]*p[i];
				return(result);
			}
		}

        public double Size()
        {
            double result = 0;
            for (int i = 0; i < this.Length; i++)
            {
                result += values[i] * values[i];
            }
            result = Math.Sqrt(result);
            return (result);
        }

        public void Normalize()
        {
            double d = this.Size();
            for (int i = 0;i<this.Length;i++)
                this[i]/=d;
        }

        /// <summary>
        /// dot product of two PointNDs
        /// </summary>
        /// <param name="p1">first argument</param>
        /// <param name="p2">secornd argument</param>
        /// <returns>dot product as Scalar</returns>
        public static Scalar dotProduct(PointND p1, PointND p2)
		{
			if (p1.Length!=p2.Length)
			{
				throw new MveException(typeof(PointND),"Cannot compute dot product of points of inequal dimension.");
			}
			else
			{
				double result = 0;
				for (int i = 0;i<p1.Length;i++) result+= p1[i]*p2[i];
				return(result);
			}
		}

		/// <summary>
		/// Get/Set element of point.
		/// </summary>
		public double this[int index]
		{
			get
			{
				if (index < this.values.Length && index >= 0)
					return this.values[index];
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.values.Length - 1) + ">");
			}
			set
			{
				if (index < this.values.Length && index >= 0)
					this.values[index] = value;
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.values.Length - 1) + ">");
			}
		} // this[int]

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string pom = "PointND (" + this.Values.Length + " dimensions): [";
			foreach (double d in Values)
				pom += d + "; ";
			pom = pom.Substring(0, pom.Length - 2);  // odstraneni prebytecneho stredniku a mezery na konci
			return pom + "]";
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(PointND) || b.GetType() != typeof(PointND))
				return false;

			PointND x = (PointND) a;
			PointND y = (PointND) b;
			if (x.Length != y.Length)
				return false;

			for (int i = 0; i < x.Length; i++)
				if (Math.Abs(x[i] - y[i]) > PointND.Epsilon)
					return false;

			return true;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return PointND.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			if (this.Length == 0)
				return 0;

			if (this.Length == 1)
				return this[0].GetHashCode();

			int hashCode = this[0].GetHashCode();
			for (int i = 1; i < this.Length; i++)
				hashCode = hashCode ^ this[i].GetHashCode();

			return hashCode;
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversion

		// ===== PointND -> Point =================
		/// <summary>
		/// Explicit conversion PointND -> Point3D
		/// </summary>
		/// <param name="p">PointND</param>
		/// <returns>Point3D</returns>
		public static explicit operator Point3D(PointND p) 
		{
			if (p.Values.Length == 1)
				return new Point3D(p.Values[0], 0, 0);
			else if (p.Values.Length == 2)
				return new Point3D(p.Values[0], p.Values[1], 0);
			else
				return new Point3D(p.Values[0], p.Values[1], p.Values[2]);
		} // PointND -> Point3D

		// ===== PointND -> Vector ================
		/// <summary>
		/// Explicit conversion PointND -> Vector2D
		/// </summary>
		/// <param name="p">PointND</param>
		/// <returns>Vector2D</returns>
		public static explicit operator Zcu.Mve.Numerics.Vector2D(PointND p) 
		{
			if (p.Values.Length == 1)
				return new Zcu.Mve.Numerics.Vector2D(p.Values[0], 0);
			else
				return new Zcu.Mve.Numerics.Vector2D(p.Values[0], p.Values[1]);
		} // PointND -> Vector2D

		/// <summary>
		/// Explicit conversion PointND -> Vector3D
		/// </summary>
		/// <param name="p">PointND</param>
		/// <returns>Vector3D</returns>
		public static explicit operator Zcu.Mve.Numerics.Vector3D(PointND p) 
		{
			if (p.Values.Length == 1)
				return new Zcu.Mve.Numerics.Vector3D(p.Values[0], 0, 0);
			else if (p.Values.Length == 2)
				return new Zcu.Mve.Numerics.Vector3D(p.Values[0], p.Values[1], 0);
			else
				return new Zcu.Mve.Numerics.Vector3D(p.Values[0], p.Values[1], p.Values[2]);
		} // PointND -> Vector3D

		/// <summary>
		/// Implicit conversion PointND -> VectorND
		/// </summary>
		/// <param name="p">PointND</param>
		/// <returns>VectorND</returns>
		public static implicit operator Zcu.Mve.Numerics.VectorND(PointND p) 
		{
			return new Zcu.Mve.Numerics.VectorND(((PointND)p.DeepCopy()).Values);
		} // PointND -> VectorND

		// ===== Vector -> PointND ================
		/// <summary>
		/// Implicit conversion Vector2D -> PointND
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>PointND</returns>
		public static implicit operator PointND(Zcu.Mve.Numerics.Vector2D v) 
		{
			return new PointND(new double[] {v.X, v.Y});
		} // Vector2D -> PointND

		/// <summary>
		/// Implicit conversion Vector3D -> PointND
		/// </summary>
		/// <param name="v">Vector3D</param>
		/// <returns>PointND</returns>
		public static implicit operator PointND(Zcu.Mve.Numerics.Vector3D v) 
		{
			return new PointND(new double[] {v.X, v.Y, v.Z});
		} // Vector3D -> PointND

		/// <summary>
		/// Implicit conversion VectorND -> PointND
		/// </summary>
		/// <param name="v">VectorND</param>
		/// <returns>PointND</returns>
		public static implicit operator PointND(Zcu.Mve.Numerics.VectorND v) 
		{
			return new PointND(((Zcu.Mve.Numerics.VectorND)v.DeepCopy()).Values);
		} // VectorND -> PointND

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(PointND a, PointND b)
		{
			return PointND.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(PointND a, PointND b) 
		{
			return !(a == b);
		}

		/// <summary>
		/// Addition operator. Addition is performed per element, when points of inequal dimension are passed an exception is raised.
		/// </summary>
		/// <param name="a">First operand</param>
		/// <param name="b">Second operand</param>
		/// <returns>Sum</returns>
		public static PointND operator +(PointND a, PointND b)
		{
			if (a.Length!=b.Length)
				throw new MveException(typeof(PointND),"Adding points of unequal dimension.");
			PointND result = new PointND(a.Length);
			for (int i = 0;i<a.Length;i++)
			{
				result[i] = a[i]+b[i];
			}
			return(result);
		}

        public static PointND crossProduct(PointND a, PointND b, PointND c)
        {
            PointND result = new PointND(4);
            result[0] = (a[1] * b[2] * c[3] + b[1] * c[2] * a[3] + c[1] * a[2] * b[3] - c[1] * b[2] * a[3] - b[1] * a[2] * c[3] - a[1] * c[2] * b[3]);
            result[1] = -(a[0] * b[2] * c[3] + b[0] * c[2] * a[3] + c[0] * a[2] * b[3] - c[0] * b[2] * a[3] - b[0] * a[2] * c[3] - a[0] * c[2] * b[3]);
            result[2] = (a[0] * b[1] * c[3] + b[0] * c[1] * a[3] + c[0] * a[1] * b[3] - c[0] * b[1] * a[3] - b[0] * a[1] * c[3] - a[0] * c[1] * b[3]);
            result[3] = -(a[0] * b[1] * c[2] + b[0] * c[1] * a[2] + c[0] * a[1] * b[2] - c[0] * b[1] * a[2] - b[0] * a[1] * c[2] - a[0] * c[1] * b[2]);
            return (result);
        }

		/// <summary>
		/// Subtraction operator. Addition is performed per element, when points of inequal dimension are passed an exception is raised.
		/// </summary>
		/// <param name="a">First operand</param>
		/// <param name="b">Second operand</param>
		/// <returns>Difference</returns>
		public static PointND operator -(PointND a, PointND b)
		{
			if (a.Length!=b.Length)
				throw new MveException(typeof(PointND),"Adding points of unequal dimension.");
			PointND result = new PointND(a.Length);
			for (int i = 0;i<a.Length;i++)
			{
				result[i] = a[i]-b[i];
			}
			return(result);
		}

		/// <summary>
		/// Left Scalar multiplication operator.
		/// </summary>
		/// <param name="k">multiplier</param>
		/// <param name="a">Point operand</param>
		/// <returns>Product</returns>
		public static PointND operator *(double k, PointND a)
		{
			PointND result = new PointND(a.Length);
			for (int i = 0;i<a.Length;i++)
			{
				result[i] = a[i]*k;
			}
			return(result);
		}

		/// <summary>
		/// Roght Scalar multiplication operator.
		/// </summary>
		/// <param name="k">multiplier</param>
		/// <param name="a">Point operand</param>
		/// <returns>Product</returns>
		public static PointND operator *(PointND a, double k)
		{
			PointND result = new PointND(a.Length);
			for (int i = 0;i<a.Length;i++)
			{
				result[i] = a[i]*k;
			}
			return(result);
		}

		#endregion

	} // class PointND
} // namespace
