using System;
using Zcu.Mve.Core;
using System.Xml;
using System.Collections;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Represents unclosed PolyLine defined by array of point indices.
	/// </summary>
	public struct PolyLine : ICell
	{
		/// <summary>
		/// Array of point indices.
		/// </summary>
		private int [] Points;

		/// <summary>
		/// Creates a PolyLine object.
		/// </summary>
		/// <param name="points">Points indices.</param>
		public PolyLine(params int[] points)
		{
			this.Points = (int []) points.Clone();
		} // PolyLine()

		/// <summary>
		/// Get the count of polyLine points.
		/// </summary>
		public int PointCount
		{
			get
			{
				if (this.Points == null)
					return 0;
				else
					return this.Points.Length;
			}
		} // PointCount

		/// <summary>
		/// Returns the point (its index in the DataSet) according to index in the PolyLine (0, 1, 2, ...).
		/// </summary>
		/// <param name="index">Index of the point in the Polygon.</param>
		/// <returns>Index of the point in the DataSet.</returns>
		public int GetPoint(int index)
		{
			return this[index];
		} // GetPoint()

		/// <summary>
		/// Get/set the concrete vertex of PolyLine. First vertex has the index 0.
		/// </summary>
		public int this[int index]
		{
			get
			{
				if (index < this.Points.Length && index >= 0)
					return (int) this.Points[index];
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.PointCount - 1) + ">");
			}
			set
			{
				if (index < this.Points.Length && index >= 0)
					this.Points[index] = value;
				else
					throw new MveException(this, "Index " + index + " is out of range <0, " + (this.PointCount - 1) + ">");
			}
		} // this[int]

		/// <summary>
		/// Method load data from Xml file to structures of curent data object.
		/// </summary>
		/// <param name="xmlTextReader">XML stram to read</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string[] points = xmlTextReader.ReadString().Trim().Split(' ', '\t');
			this.Points = new int [points.Length];

			for (int i = 0; i < points.Length; i++)
				this.Points[i] = Int32.Parse(points[i].Trim());
		} // ReadData()

		/// <summary>
		/// Method write data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Data stream to write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			string pom = "";
			for (int i = 0; i < this.PointCount; i++)
				pom += this.Points[i].ToString() + " ";
			pom = pom.Substring(0, pom.Length - 1);	 // odstraneni prebytecne mezery na konci
			xmlTextWriter.WriteString(pom);
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>same data, different memmory space.</returns>
		public IDataObject DeepCopy()
		{
			PolyLine tmp = new PolyLine();
			tmp.Points = (int []) this.Points.Clone();
			return tmp;
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// Returns false if some index is negative.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			if (this.PointCount == 0)
			{
				Console.WriteLine("Error! " + this.GetType().Name + ": contains no indices.");
				return false;
			}

			for (int i = 0; i < this.PointCount; i++)
				if (this[i] < 0) 
				{
					Console.WriteLine("Error! " + this.GetType().Name + ": point No. " + i + " has negative index.");
					return false;
				}
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string pom = "PolyLine (" + this.PointCount + " points): {";
			if (this.PointCount == 0)
				return pom += "}";
			foreach (int index in this.Points)
				pom += index + ", ";
			return pom.Substring(0, pom.Length - 2) + "}";
		} // ToString()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

	} // class PolyLine
} // namespace
