using System;
using Zcu.Mve.Core;
using System.Collections;
using System.Xml;
using System.Globalization;
using System.Diagnostics;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// This class represents a structured, semiregular, 2-dimensional grid. <br />
	/// It means that topology is given implicitly and geometry is given by two arrays of scalar values.
	/// </summary>
	[DataObject("Structured, semiregular, 2-dimensional grid.")]
	public class RectilinearGrid2D : Grid2D
	{
		/// <summary>
		/// Field represents x-coordinates of points.
		/// Values have to be between 0 and 1.
		/// </summary>
		private double[] arrayWidth;

		/// <summary>
		/// Field represents y-coordinates of points.
		/// Values have to be between 0 and 1.
		/// </summary>
		private double[] arrayHeight;

		/// <summary>
		/// Creates an epmty object (Used by XmlLoader, otherwise useless).
		/// </summary>
		public RectilinearGrid2D()
		{
			this.arrayWidth = null;
			this.arrayHeight = null;
		} // RectilinearGrid2D()

		/// <summary>
		/// Constructor - creates empty grid.
		/// </summary>
		/// <param name="arrayHeight">Field represents y-coordinates of points.</param>
		/// <param name="arrayWidth">Field represents x-coordinates of points.</param>
		public RectilinearGrid2D(double[] arrayHeight, double[] arrayWidth) : base(arrayHeight.Length-1, arrayWidth.Length-1)
		{
			this.arrayHeight = (double []) arrayHeight.Clone();
			this.arrayWidth = (double []) arrayWidth.Clone();
		} // RectilinearGrid2D()

		/// <summary>
		/// Returns reference to internal array.
		/// </summary>
		public double[] ArrayHeight
		{
			get
			{
				return this.arrayHeight;
			}
		} // ArrayWidth

		/// <summary>
		/// Returns reference to internal array.
		/// </summary>
		public double[] ArrayWidth
		{
			get
			{
				return this.arrayWidth;
			}
		} // ArrayHeight

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			string str = "Visualization.RectilinearGrid2D:\r\n";
			str += "height array: ";
			foreach (double d in this.ArrayHeight)
			{
				str += " " + d.ToString(Globals.nfi);
			}
			str += "\r\nwidth array: ";
			foreach (double d in this.ArrayWidth)
			{
				str += " " + d.ToString(Globals.nfi);
			}
			str += "\r\n" + base.ToString();
      
			return str;
		}

		// ===== Methods of DataObject ========================================================
		#region Methods of DataObject

		/// <summary>
		/// Method reads data from Xml file to structures of data object.
		/// </summary>
		/// <param name="xmlTextReader">Data stream to be read.</param>
		public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			ReadHeader(xmlTextReader);

			this.arrayHeight = new double [this.Height+1];
			this.arrayWidth = new double [this.Width+1];

			while(xmlTextReader.Read())
			{
				ReadNode(xmlTextReader);
			} // while
		} // ReadData()

		/// <summary>
		/// Reads array from xml.
		/// </summary>
		/// <param name="array">Array of double values.</param>
		/// <param name="xmlTextReader">Data stream to be read.</param>
		private void ReadDoubleArray(double [] array, XmlTextReader xmlTextReader)
		{
			// TODO: Dodelat nacitani ...
			if (xmlTextReader.NodeType == XmlNodeType.Element)
			{
				string [] strs = xmlTextReader.ReadElementString().Trim().Split(' ');

				if (strs.Length != array.Length)
				{
					throw new MveException(this, "Length of coordinate array do not agree with declared size of grid.");
				}

				for (int i = 0; i < strs.Length; i++)
				{
					array[i] = Double.Parse(strs[i], Globals.nfi);
				}
			}
		}

		/// <summary>
		/// Reads pointattrs and cellattrs nodes.
		/// </summary>
		/// <param name="xmlTextReader">Data stream to be read.</param>
		protected void ReadNode(XmlTextReader xmlTextReader)
		{
			if (xmlTextReader.NodeType == XmlNodeType.Element)
			{
				switch(xmlTextReader.Name)
				{
					case "width-array":	
						this.ReadDoubleArray(this.ArrayWidth, xmlTextReader);
						break;
					case "height-array":
						this.ReadDoubleArray(this.ArrayHeight, xmlTextReader);
						break;
					default:
						base.ReadXmlNode(xmlTextReader);
						break;
				} // switch
			}      
		}

		/// <summary>
		/// Support method for writing a double array into xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Data stream to write.</param>
		/// <param name="array">Array of double values.</param>
		/// <param name="elemName">Name of xml element.</param>
		private void WriteArray(XmlTextWriter xmlTextWriter, double[] array, string elemName)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteStartElement(elemName);
			string tmp = "";
			for (int i = 0; i < array.Length; i++)
				tmp += array[i].ToString(nfi) + " ";
			xmlTextWriter.WriteString(tmp.Substring(0, tmp.Length - 1));
			xmlTextWriter.WriteEndElement();
		} // WriteArray()


		/// <summary>
		/// Write arrays of x and y coordinates.
		/// </summary>
		/// <param name="xmlTextWriter">Data stream to write.</param>
		protected void WriteArrays(XmlTextWriter xmlTextWriter)
		{
			WriteArray(xmlTextWriter, this.arrayWidth, "width-array");
			WriteArray(xmlTextWriter, this.arrayHeight, "height-array");
		}

		/// <summary>
		/// Method write content of this structure into Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Data stream to write.</param>
		public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			base.WriteHeader(xmlTextWriter);
			this.WriteArrays(xmlTextWriter);
			base.WritePointAttrs(xmlTextWriter);
			base.WriteCellAttrs(xmlTextWriter);
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public override IDataObject DeepCopy()
		{
			RectilinearGrid2D copy = (RectilinearGrid2D) base.DeepCopy();
      
			copy.arrayWidth = (double []) this.ArrayWidth.Clone();
			copy.arrayHeight = (double []) this.ArrayHeight.Clone();

			return copy;
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public override bool CheckConsistence()
		{
			bool ret = true;
      
			for (int i = 1; i < this.Height; i++)
			{
				if (this.arrayHeight[i-1] >= this.arrayHeight[i])
				{
					Console.WriteLine("X-values do not preserve ordinality. Index: " + i + ".");
				}
			}

			for (int j = 1; j < this.Width; j++)
			{
				if (this.arrayWidth[j-1] >= this.arrayWidth[j])
				{
					Console.WriteLine("X-values do not preserve ordinality. Index: " + j + ".");
				}
			}

			return ret && base.CheckConsistence();
		} // CheckConsistence()

		#endregion

		// ===== Methods of Grid2D ========================================================
		#region Methods of Grid2D

		/// <summary>
		/// Returns x, y coordinates of point with i, j indices.
		/// </summary>
		/// <param name="i">Row index of point.</param>
		/// <param name="j">Column index of point.</param>
		/// <returns>New  instance of point.</returns>
		public override Point2D GetPoint(int i, int j)
		{
			if (i > (this.Height + 1) || j > (this.Width + 1) || i < 0 || j < 0)
				throw new MveException(this, this.GetType().Name + ": point [" + i + ", " + j + "] is out of range \n" + this.ToString());

			return new Point2D(this.arrayWidth[j], this.arrayHeight[i]);
		} // GetPoint()

		#endregion

	} // class RectilinearGrid2D
} // namespace

