using System;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// A module that converts a Scalar attribute to ColorRGBA.
    /// It first samples a palette by asking for colors by setting the output "Palette value" to numbers in the (0,1) interval.
    /// Then it assigns the closest color to the points. Point with lowest attribute value will receive color returned for the 0 value,
    /// point with largest attribute value will receive color returned when 1 was set to the "Palette value",
    /// i.e. the whole palette is used for the scale of the attribute.
    /// </summary>
    [ModuleInfo("Libor Va", "A module that converts a Scalar attribute to ColorRGBA."
    +"It first samples a palette by asking for colors by setting the output 'Palette value' to numbers in the (0,1) interval."
    +"Then it assigns the closest color to the points. Point with lowest attribute value will receive color returned for the 0 value,"
    +"point with largest attribute value will receive color returned when 1 was set to the 'Palette value',"
    +"i.e. the whole palette is used for the scale of the attribute.")]
	[PortInfo("UnstrGrid", "A grid to be processed.")]
	[PortInfo("Palette input", "A port where a color is expected when a value is set to the Palette value output port.")]
	[PortInfo("Palette value", "A poert where values form the range 0-1 are put and colors are sampled from the Palette value port.")]
    public class ScalarAttr2ColorTransformer: Module
    {
        private bool outset = false;

		private UnstrGrid output = null;

        private string attributeName = "";

        /// <summary>
        /// Name of the attribut to be converted to ColorRGBA.
        /// If set to an empty string, then the first point attribute of the Scalar type is used.
        /// </summary>
        public string AttributeName
        {
            get { return (attributeName); }
            set
            {
                if (value == null)
                {
                    throw new MveException(this, "Null passed as attribut name.");
                }                

                attributeName = value;
            }
        }

        private bool resamplePalette = false;

		/// <summary>
		/// Determines wether the palette shoud be resampled each time or not.
		/// </summary>
		[Description("Determines wether the palette shoud be resampled each time or not.")]
        public bool ResamplePalette
        {
            get { return (resamplePalette); }
            set { resamplePalette = value; }
        }

		/// <summary>
		/// default constructor. Output port is not added, because it is created in runtime when an input is attached. This allows the output to be of the same type as the input.
		/// </summary>
        public ScalarAttr2ColorTransformer()
        {
            AddInPort("UnstrGrid", typeof(UnstrGrid));
            AddInPort("Palette input", typeof(ColorRGBA), true, false);
            AddOutPort("Palette value", typeof(Scalar), false);

            this.OnConnectEvent += new ConnectCallBack(ScalarAttr2ColorTransformer_OnConnectEvent);
            this.OnDisconnectEvent += new DisconnectCallBack(ScalarAttr2ColorTransformer_OnDisconnectEvent);
        }

        void ScalarAttr2ColorTransformer_OnDisconnectEvent(object source, InPort portTo, out bool allowed)
        {
            if (portTo == this.GetInPort("UnstrGrid"))
            {
                if (outset)
                {
                    RemoveOutPort("Output");
                    outset = false;
                }
            }
            allowed = true;
        }

        void ScalarAttr2ColorTransformer_OnConnectEvent(object source, OutPort portFrom, InPort portTo, out bool allowed)
        {
            if (portTo == this.GetInPort("UnstrGrid"))
            {
                this.AddOutPort("Output", portFrom.PortType);
                outset = true;
            }
            allowed = true;
        }

        private int stepCount = 500;

		/// <summary>
		/// Number of steps of the internal sampled palette
		/// </summary>
		[Description("Number of steps of the internal sampled palette")]
        public int StepCount
        {
            get { return (stepCount); }
            set
            {
                if (value > 0)
                {
                    stepCount = value;
                }
                else
                {
                    throw new MveException(this, "Number of palette steps cannot be less than 1.");
                }
            }
        }

        private ColorRGBA[] palette;

		/// <summary>
		/// Performs the sampling of the palette and the conversion of the attribute to colors.
		/// </summary>
        public override void Execute()
        {
            if ((resamplePalette)||(palette == null))
            {
                palette = new ColorRGBA[stepCount];                
                for (int i = 0; i < stepCount; i++)
                {
                    SetOutput("Palette value",new Scalar(i*(1.0/stepCount)));
                    UpdateInput("Palette input");
                    palette[i] = (ColorRGBA)GetInput("Palette input");
                }
            }

            if (!IsDataSame("UnstrGrid")||resamplePalette||(output == null))
            {
                UnstrGrid input = (UnstrGrid)GetInput("UnstrGrid");

                if (attributeName == "")
                {
                    foreach (string s in input.GetPointAttrsIds())
                    {
                        if (input.GetPointAttrs(s).ElemType == typeof(Scalar))
                            attributeName = s;
                    }
                }

                UniformDataArray attribute = input.GetPointAttrs(attributeName);

                double min = double.MaxValue;
                double max = double.MinValue;
                double v;
                     
                for (int i = 0; i < attribute.Count; i++)
                {
                    v = ((Scalar)(attribute[i])).Value;
                    if (v < min) min = v;
                    if (v > max) max = v;
                }

                output = input.ShallowCopy();

                UniformDataArray colors = new UniformDataArray(typeof(ColorRGBA), input.Points.Count);
                for (int i = 0; i < input.Points.Count; i++)
                {
                    v = ((Scalar)attribute[i]).Value;
                    int index = (int)((stepCount*(v-min))/(max-min));
                    if (index<0) index = 0;
                    if (index>=stepCount) index = stepCount-1;
                    colors[i] = palette[index];
                    if (index < 10)
                    {
                        index = index * 1;
                    }
                }

                output.AddPointAttrs(attributeName + "Colors", colors);
                SetOutput("Output", output);
            }
            else
            {
                SetOutput("Output", output, true);
            }
        }
    }
}
