using System;
using Zcu.Mve.Core;
using System.Xml;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Class represent topology of a Tetrahedron. Defined by four point indices.
	/// </summary>
	public struct Tetrahedron : ICell
	{
		/// <summary>
		/// Index of first vertex of the Tetrahedron.
		/// </summary>
		public int V1;
		/// <summary>
		/// Index of second vertex of the Tetrahedron.
		/// </summary>
		public int V2;
		/// <summary>
		/// Index of third vertex of the Tetrahedron.
		/// </summary>
		public int V3;
		/// <summary>
		/// Index of fourth vertex of the Tetrahedron.
		/// </summary>
		public int V4;

		/// <summary>
		/// Creates a Tetrahedron object.
		/// </summary>
		/// <param name="V1">Index of first vertex of the Tetrahedron.</param>
		/// <param name="V2">Index of second vertex of the Tetrahedron.</param>
		/// <param name="V3">Index of third vertex of the Tetrahedron.</param>
		/// <param name="V4">Index of fourth vertex of the Tetrahedron.</param>
		public Tetrahedron(int V1, int V2, int V3, int V4)
		{
			this.V1 = V1;
			this.V2 = V2;
			this.V3 = V3;
			this.V4 = V4;
		} // Triangle()

		/// <summary>
		/// Returns array of three vertices of the Tetrahedron.
		/// Change of returned array do not cause change inside line object.
		/// </summary>
		/// <remarks>This access is not efficient. Use V1, V2, V3, V4 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2, V3, V4 instead.", false)]
		public int [] PointArray
		{
			get
			{
				return new int [4] {V1, V2, V3, V4};
			}
		} // PointArray

		/// <summary>
		/// Return number of points of cell.
		/// </summary>
		public int PointCount
		{
			get
			{
				return 4;
			}
		} // PointCount

		/// <summary>
		/// Returns index of vertex.
		/// </summary>
		/// <param name="index">0 .. 3 are only acceptable vaules.</param>
		/// <returns>Index of vertex.</returns>
		/// <remarks>This access is not efficient. Use V1, V2, V3, V4 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2, V3, V4 instead.", false)]
		public int GetPoint(int index)
		{
			switch (index)
			{
				case 0:
					return V1;
				case 1:
					return V2;
				case 2:
					return V3;
				case 3:
					return V4;
				default:
					throw new MveException(this, "Index out of range. Tetrahedron has only four vertices.");
			}
		} // GetPoint()

		/// <summary>
		/// Get/Set index of Tetrahedron vertex
		/// </summary>
		/// <remarks>This access is not efficient. Use V1, V2, V3, V4 instead.</remarks>
		[ObsoleteAttribute("This access is not efficient. Use V1, V2, V3, V4 instead.", false)]
		public int this[int index]
		{
			get
			{
				switch (index)
				{
					case 0:
						return V1;
					case 1:
						return V2;
					case 2:
						return V3;
					case 3:
						return V4;
					default:
						throw new MveException(this, "Index out of range. Triangle have only three vertices.");
				}					
			}
			set
			{
				switch (index)
				{
					case 0:
						V1 = value;
						break;
					case 1:
						V2 = value;
						break;
					case 2:
						V3 = value;
						break;
					case 3:
						V4 = value;
						break;
					default:
						throw new MveException(this, "Index out of range. Triangle have only three vertices.");
				}						
			}
		} // this[int]

		/// <summary>
		/// Read data from XML file.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string [] vertices = xmlTextReader.ReadString().Trim().Split(' ', '\t');

			if (vertices.Length != 4)
				throw new MveException(this, "Incorect number of vertices in tetrahedron.");

			this.V1 = Int32.Parse(vertices[0].Trim());
			this.V2 = Int32.Parse(vertices[1].Trim());
			this.V3 = Int32.Parse(vertices[2].Trim());
			this.V4 = Int32.Parse(vertices[3].Trim());
		} // ReadData()

		/// <summary>
		/// Write data to XML file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteString(V1 + " " + V2 + " " + V3 + " " + V4);
		} // WriteData()

		/// <summary>
		/// Deep copy of curent object.
		/// </summary>
		/// <returns>Deep copy (same data, different memory place).</returns>
		public IDataObject DeepCopy()
		{
			return new Tetrahedron(this.V1, this.V2, this.V3, this.V4);
		} // DeepCopy()

		/// <summary>
		/// Tetrahedron with negative index is considered inconsistent.
		/// If two, three or four identical indices are detected warning is given.
		/// </summary>
		/// <returns>True if the Tetrahedron is consistent.</returns>
		public bool CheckConsistence()
		{
			if (this.V1 < 0 || this.V2 < 0 || this.V3 < 0 || this.V4 < 0)
			{
				Console.WriteLine("Error! " + this.GetType().Name + " has negative index.");
				return false;
			}
			if (this.V1 ==  this.V2 && this.V2 == this.V3 && this.V3 == this.V4)
				Console.WriteLine("Warning! " + this.GetType().Name + " deformation: all 4 points have identical index.");
			else if (this.V1 == this.V2 && this.V2 == this.V3 ||
				this.V1 == this.V2 && this.V2 == this.V4 ||
				this.V1 == this.V3 && this.V3 == this.V4 ||
				this.V2 == this.V3 && this.V3 == this.V4)
				Console.WriteLine("Warning! " + this.GetType().Name + " deformation: some 3 points have identical index.");
			else if (this.V1 == this.V2 || this.V1 == this.V3 || this.V1 == this.V4
				|| this.V2 == this.V3 || this.V2 == this.V4 
				|| this.V3 == this.V4)
				Console.WriteLine("Warning! " + this.GetType().Name + " deformation: some 2 points have identical index.");
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return String.Format("Triangle: [{0}, {1}, {2}, {3}]", V1, V2, V3, V4);
		} // ToString()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		private double det(Point3D p1, Point3D p2, Point3D p3, Point3D p4)
		{
			double result = 0;
			result+= p1.X*p2.Y*p3.Z -p1.X*p2.Y*p4.Z +p1.X*p2.Z*p4.Y -p1.X*p2.Z*p3.Y +p1.X*p3.Y*p4.Z -p1.X*p3.Z*p4.Y;
			result+=-p1.Y*p2.Z*p4.X +p1.Y*p2.Z*p3.X -p1.Y*p3.X*p4.Z +p1.Y*p3.Z*p4.X -p1.Y*p2.X*p3.Z +p1.Y*p2.X*p4.Z;
			result+= p1.Z*p3.X*p4.Y -p1.Z*p3.Y*p4.X +p1.Z*p2.X*p3.Y -p1.Z*p2.X*p4.Y +p1.Z*p2.Y*p4.X -p1.Z*p2.Y*p3.X;
			result+=-p2.X*p3.Y*p4.Z +p2.X*p3.Z*p4.Y -p2.Y*p3.Z*p4.X +p2.Y*p3.X*p4.Z -p2.Z*p3.X*p4.Y +p2.Z*p3.Y*p4.X;
			return(result);
		}

		/// <summary>
		/// determines wether a point3D lies within a the tetrahedron.
		/// </summary>
		/// <param name="p">The Point3D in question</param>
		/// <param name="pointsCoordinates">UniformDataArray of vertex coordinates.</param>
		/// <returns>true if the point lies within the tetrahedron, false otherwise</returns>
		public bool ContainsPoint3D(Point3D p, UniformDataArray pointsCoordinates)
		{
			Point3D p1 = (Point3D)pointsCoordinates[V1];
			Point3D p2 = (Point3D)pointsCoordinates[V2];
			Point3D p3 = (Point3D)pointsCoordinates[V3];
			Point3D p4 = (Point3D)pointsCoordinates[V4];

			double d0 = det(p1,p2,p3,p4);
			double d1 = det(p,p2,p3,p4);
			double d2 = det(p1,p,p3,p4);
			double d3 = det(p1,p2,p,p4);
			double d4 = det(p1,p2,p3,p);

			if (((d1/d0)>0)&&((d2/d0)>0)&&((d3/d0)>0)&&((d4/d0)>0))
			{
				return(true);
			}
			else return(false);

		}

		/// <summary>
		/// Equality test based on vertex index equality. Order of vertices must be equal in order for the test to return true.
		/// </summary>
		/// <param name="obj">Target object to be tested. Only instances of Tetrahedron or it's descendant may return true, other data types return false</param>
		/// <returns>true if vertex indexes are equal, false otherwise</returns>
		public override bool Equals(object obj)
		{
			if (obj is Tetrahedron)
			{
				Tetrahedron t = (Tetrahedron)obj;
				if ((t.V1 == this.V1)&&(t.V2 == this.V2)&&(t.V3 == this.V3)&&(t.V4 == this.V4))
				{
					return(true);
				}
			}
			return(false);
		}

        public override int GetHashCode()
        {
            return (this.V1 ^ this.V2*100 ^ this.V3*10000 ^ this.V4*1000000);
        }

	} // class Tetrahedron
} // namespace
