using System;
using Zcu.Mve.Core;
using System.Globalization;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// This structure represents Triangle neighbors.
	/// </summary>
	public struct TriangleNeighbor : INeighbor
	{
		/// <summary>
		/// Indices to three neighboring triangles.
		/// </summary>
		public int [] Neighbours;

		/// <summary>
		/// Number of neighbors
		/// </summary>
		public int Count;

		/// <summary>
		/// Creates an object with values .
		/// </summary>
		/// <param name="t1">Index of neighbor.</param>
		/// <param name="t2">Index of neighbor.</param>
		/// <param name="t3">Index of neighbor.</param>
		/// <param name="count">Number of neighbors.</param>
		public TriangleNeighbor(int t1, int t2, int t3, int count)
		{
			this.Neighbours = new int[] {t1,t2,t3};    
			this.Count = count;
		} // TriNeighbor()

		/// <summary>
		/// Creates an object with values.
		/// </summary>
		/// <param name="t1">Index of neighbor.</param>
		/// <param name="t2">Index of neighbor.</param>
		/// <param name="t3">Index of neighbor.</param>
		public TriangleNeighbor(int t1, int t2, int t3)
		{
			this.Neighbours = new int[] {t1,t2,t3};
      
			int count = 0;
			foreach (int index in this.Neighbours)
				if (index >= 0)
					count++;

			this.Count = count;
		} // TriNeighbor()

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] split = xmlTextReader.ReadString().Trim().Split(' ', '\t');
      
			if (split.Length != 4)
				throw new MveException(this, "Invalid TriangleNeighbour data format.");

			this.Neighbours[0] = Int32.Parse(split[0], nfi);
			this.Neighbours[1] = Int32.Parse(split[1], nfi);
			this.Neighbours[2] = Int32.Parse(split[2], nfi);
			this.Count = Int32.Parse(split[3], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteString(this.Neighbours[0].ToString(nfi) + " " + 
				this.Neighbours[1].ToString(nfi) + " " + 
				this.Neighbours[2].ToString(nfi) + " " + 
				this.Count.ToString(nfi));
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new TriangleNeighbor(this.Neighbours[0], this.Neighbours[1], this.Neighbours[2], this.Count);
		} // DeepCopy()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return "TriangleNeighbor (" + this.Count + ") : " + this.Neighbours[0] + "; " + this.Neighbours[1] +  "; " + this.Neighbours[2] + "]";
		} // ToString()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			if (this.Neighbours == null)
			{
				Console.WriteLine("TriangleNeighbor: Neighbours array is null.");
				return false;
			}
			if (this.Neighbours.Length != 3)
			{
				Console.WriteLine("TriangleNeighbor have incorrect number of neighbours.");
				return false;
			}
			int i = 0;
			foreach (int index in this.Neighbours)
			{
				if (index >= 0)
				{
					i++;
				}
			}
			if (i != this.Count)
			{
				Console.WriteLine("Number of valid indices do not match Count value.");
				return false;
			}

			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

	} // TriangleNeighbor
} // namespace
