using System;
using System.Collections;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Specialization of  UnstrGrid.
	/// It contains a hash table that represents a set of available cell types.
	/// For more details see Mesh class.
	/// </summary>
	[DataObject("Specialization of  UnstrGrid. It contains a hash table that represents a set of available cell types. For more details see Mesh class.")]
	public abstract class UnstrGridEigCell : Zcu.Mve.Visualization.UnstrGrid
	{
		/// <summary>
		/// Contains System.Type as keys. 
		/// It represents type of cells allowed for current UnstrGrid 
		/// (i.e. Mesh, StripMesh, TriangleMesh...).
		/// </summary>
		protected Hashtable eigCells = new Hashtable();

		/// <summary>
		/// Implicit constructor.
		/// </summary>
		public UnstrGridEigCell()
		{
		}

		// MFR: Do not required ...
		/// <summary>
		/// Callback method for array that check whether particular cell is eigen cell for current mesh.
		/// </summary>
		/// <param name="index"></param>
		/// <param name="element"></param>
		/// <returns></returns>
		private bool ElementSetCheck(int index, DataObject element)
		{
			if (!eigCells.ContainsKey(element.GetType()))
			{
				throw new MveException(this, element.GetType().FullName + " is not eigen cell for mesh where curent array is a cell array.");
			}
			return true;
		}

		/// <summary>
		/// Method add cell array and check whether type of cells is in eigCells array.
		/// </summary>
		/// <param name="cellId">ID of the cell.</param>
		/// <param name="cells">Uniform data array of cells.</param>
		public override void AddCells(string cellId, UniformDataArray cells)
		{
			if (!eigCells.ContainsKey(cells.ElemType))
			{
				throw new MveException(this, "Given array of cell contains non eigen cells for UnstrGridEigCell (" + cells.ElemType.FullName + " is not eigen cells).");
			}
			// cells.OnElementSetEvent += new AbstractDataArray.ElementSetCallBack(ElementSetCheck);
			base.AddCells(cellId, cells);
		}

		/// <summary>
		/// Creates deep copy of current UnstrGridEigCell (i.e. Mesh, TriangleMesh...).
		/// </summary>
		/// <returns>Deep copy of current UnstrGridEigCell (same data, different memory space).</returns>
		public override IDataObject DeepCopy()
		{
			UnstrGridEigCell ret = (UnstrGridEigCell) Activator.CreateInstance(this.GetType());
			
			// copy transformation
			if (this.Transform != null)
				ret.Transform = this.Transform;
			
			// copy points
			if (this.Points != null)
				ret.Points = (UniformDataArray) this.Points.DeepCopy();

			// copy libnames
			ret.libNames.AddRange(this.libNames);

			// copy cells
			foreach (string cellId in this.GetCellIds())
			{
				UniformDataArray newCells = (UniformDataArray) ((UniformDataArray) this.GetCells(cellId)).DeepCopy();
				ret.AddCells(cellId, newCells);
				
				// copy cell attributes
				foreach (string attrId in this.GetCellAttrsIds(cellId))
				{
					UniformDataArray newAttrs = (UniformDataArray) ((UniformDataArray) this.GetCellAttrs(cellId, attrId)).DeepCopy();
					ret.AddCellAttrs(cellId, attrId, newAttrs);
				}
			}

			// copy point attributes
			foreach (string attrId in this.GetPointAttrsIds())
			{
				UniformDataArray newAttrs = (UniformDataArray) ((UniformDataArray) this.GetPointAttrs(attrId)).DeepCopy();
				ret.AddPointAttrs(attrId, newAttrs);
			}
			return ret;
		}

	} // class UnstrGridEigCell
} // namespace